#include "huffman.h"
#include "compress.h"
unsigned char * rotcmpbuf;
int rotcmpbufsize;

int rotcmp(const void *l, const void *r);

int compress(unsigned char * inbuf, unsigned char * outbuf, unsigned int size, code CODES[256], int * n) {
    int index;
    unsigned char * tempbuf = (unsigned char *) malloc(size * sizeof(char)); //create temporary buffer
    unsigned char * tempbuf2 = (unsigned char *) malloc(size * sizeof(char)); //create temporary buffer
    index = burrows_wheeler_encode(inbuf, tempbuf, size); //apply BWT
    mtf_encode(tempbuf, tempbuf2, size); //apply MTF
    *n = huffman(tempbuf2, outbuf, size, CODES); //apply Huffman encoding
    free(tempbuf);
    free(tempbuf2);
    return index;
}

int decompress(unsigned char * inbuf, unsigned char * outbuf, unsigned int size, int index, code CODES[256]) {
    int n;
    unsigned char * tempbuf = (unsigned char *) malloc(size * sizeof(char)); //temporary buffer
    unsigned char * tempbuf2 = (unsigned char *) malloc(size * sizeof(char)); //temporary buffer    
    n = get_bits(inbuf, tempbuf, CODES, size);
    mtf_decode(tempbuf, tempbuf2, n); //decode MTF
    burrows_wheeler_decode(tempbuf2, outbuf, n, index); //decode BWT
    free(tempbuf);
    free(tempbuf2);
    return n;
}

int burrows_wheeler_encode(unsigned char * inbuf, unsigned char * outbuf, unsigned int size) {
    int indices[size];
    int i;

    for(i=0; i<size; i++)
        indices[i] = i; //initialize
    rotcmpbuf = inbuf;
    rotcmpbufsize = size;
    qsort (indices, size, sizeof(int), rotcmp); //sort the table

    for (i=0; i<size; i++)
        outbuf[i] = inbuf[(indices[i]+size-1)%size];
    for (i=0; i<size; i++)
    {
        if (indices[i] == 1) {
            return i;
        }
    }
}

int rotcmp(const void *l, const void *r) //we don't create real table for sorting, just calculate rotations for comparisions
{
    int li = *(const int*)l, ri = *(const int*)r, ac=rotcmpbufsize;
    while (rotcmpbuf[li] == rotcmpbuf[ri])
    {
        if (++li == rotcmpbufsize)
            li = 0;
        if (++ri == rotcmpbufsize)
            ri = 0;
        if (!--ac)
            return 0;
    }
    if (rotcmpbuf[li] > rotcmpbuf[ri])
        return 1;
    else
        return -1;
}

int burrows_wheeler_decode(unsigned char * inbuf, unsigned char * outbuf, unsigned int size, int index)
{
    unsigned char F[size];
    int buckets[256];
    int i,j,k;
    int indices[size];

    for (i=0; i<256; i++)
        buckets[i] = 0;
    for (i=0; i<size; i++)
        buckets[inbuf[i]] ++;
    for (i=0,k=0; i<256; i++)
        for (j=0; j<buckets[i]; j++)
            F[k++] = i;
    
    for (i=0,j=0; i<256; i++)
    {
        while (i>F[j] && j<size)
            j++;
        buckets[i] = j;
    }
    for(i=0; i<size; i++)
        indices[buckets[inbuf[i]]++] = i;
    for(i=0,j=index; i<size; i++)
    {
        outbuf[i] = inbuf[j];
        j=indices[j];
    }
}

void mtf_encode(unsigned char * inbuf, unsigned char * outbuf, int size) //MTF encoding
{
    int i,j;
    unsigned char state[256];

    for(i=0; i <256; i++)
        state[i] = i;
    for (i=0; i<size; i++) {
        outbuf[i] = state[inbuf[i]];
        for (j=0; j<256; j++)
            if (state[j] < state[inbuf[i]])
                state[j] ++;
        state[inbuf[i]] = 0;
    }
}

void mtf_decode(unsigned char * inbuf, unsigned char * outbuf, int size) //MTF decoding
{
    int i,j;
    unsigned char state[256];
    unsigned char tmp;

    for (i=0; i<256; i++)
        state[i] = i;
    for (i=0; i<size; i++) {
        outbuf[i] = state[inbuf[i]];
        tmp = state[inbuf[i]];
        for (j=inbuf[i]; j ; j--)
            state[j] = state[j-1];
        state[0] = tmp;
    }
}
